;------------------------------------------------------------------------------
;-         ATMEL Microcontroller Software Support  -  ROUSSET  -
;------------------------------------------------------------------------------
; The software is delivered "AS IS" without warranty or condition of any
; kind, either express, implied or statutory. This includes without
; limitation any warranty or condition with respect to merchantability or
; fitness for any particular purpose, or against the infringements of
; intellectual property rights of others.
;-----------------------------------------------------------------------------
;- File source          : Cstartup.s
;- Object               : Generic CStartup for KEIL and SAM7X256
;- Compilation flag     : None
;-
;- 1.0 12/Sep/05 JPP    : Creation
;------------------------------------------------------------------------------

/* 
//*** <<< Use Configuration Wizard in Context Menu >>> *** 
*/

// *** Startup Code (executed after Reset) ***


#include		"include/AT91SAM7X256_inc.h"

;--------------------------------
;- ARM Core Mode and Status Bits
;--------------------------------
ARM_MODE_USER           EQU     0x10
ARM_MODE_FIQ            EQU     0x11
ARM_MODE_IRQ            EQU     0x12
ARM_MODE_SVC            EQU     0x13
ARM_MODE_ABORT          EQU     0x17
ARM_MODE_UNDEF          EQU     0x1B
ARM_MODE_SYS            EQU     0x1F

I_BIT                   EQU     0x80
F_BIT                   EQU     0x40
T_BIT                   EQU     0x20

/*
// <h> Stack Configuration (Stack Sizes in Bytes)
//   <o0> Undefined Mode      <0x0-0xFFFFFFFF>
//   <o1> Supervisor Mode     <0x0-0xFFFFFFFF>
//   <o2> Abort Mode          <0x0-0xFFFFFFFF>
//   <o3> Fast Interrupt Mode <0x0-0xFFFFFFFF>
//   <o4> Interrupt Mode      <0x0-0xFFFFFFFF>
//   <o5> Top_Stack           <0x0-0xFFFFFFFF>
// </h>
*/
        UND_Stack_Size  EQU     0x00000000
        SVC_Stack_Size  EQU     0x00000000
        ABT_Stack_Size  EQU     0x00000000
        FIQ_Stack_Size  EQU     0x00000000
        IRQ_Stack_Size  EQU     0x00000060
									
$IF (RAM)
        Top_Stack       EQU     0x00010000
$ELSE
        Top_Stack       EQU     0x00210000
$ENDIF

;------------------------------------------------------------------------------
;- Area Definition
;------------------------------------------------------------------------------
AREA   STARTUPCODE, CODE, AT 0x00000000   // READONLY, ALIGN=4
       PUBLIC  __startup

__startup       PROC    CODE32
;------------------------------------------------------------------------------
;- Exception vectors 
;--------------------
;- These vectors can be read at address 0 or at RAM address
;- They ABSOLUTELY requires to be in relative addresssing mode in order to
;- guarantee a valid jump. For the moment, all are just looping.
;- If an exception occurs before remap, this would result in an infinite loop.
;- To ensure if a exeption occurs before start application to infinite loop.
;------------------------------------------------------------------------------
                B           InitReset           ; 0x00 Reset handler
undefvec:
                B           undefvec            ; 0x04 Undefined Instruction
swivec:
                B           swivec              ; 0x08 Software Interrupt
pabtvec:
                B           pabtvec             ; 0x0C Prefetch Abort
dabtvec:
                B           dabtvec             ; 0x10 Data Abort
rsvdvec:
                B           rsvdvec             ; 0x14 reserved
irqvec:
                B           IRQ_Handler_Entry   ; 0x18 IRQ
fiqvec:               				; 0x1c FIQ
;------------------------------------------------------------------------------
;- Function             : FIQ_Handler_Entry
;- Treatments           : FIQ Controller Interrupt Handler.
;- Called Functions     : AIC_FVR[interrupt] 
;------------------------------------------------------------------------------
FIQ_Handler_Entry:  
;- Switch in SVC/User Mode to allow User Stack access for C code 
; because the FIQ is not yet acknowledged

;- Save and r0 in FIQ_Register 
            mov         r9,r0
            ldr         r0 , [r8, #AIC_FVR]
            msr         CPSR_c,#I_BIT | F_BIT | ARM_MODE_SVC

;- Save scratch/used registers and LR in User Stack
            stmfd       sp!, { r1-r3, r12, lr}

;- Branch to the routine pointed by the AIC_FVR
            mov         r14, pc
            bx          r0

;- Restore scratch/used registers and LR from User Stack
            ldmia       sp!, { r1-r3, r12, lr}

;- Leave Interrupts disabled and switch back in FIQ mode
            msr         CPSR_c, #I_BIT | F_BIT | ARM_MODE_FIQ

;- Restore the R0 ARM_MODE_SVC register 
            mov         r0,r9

;- Restore the Program Counter using the LR_fiq directly in the PC
            subs        pc,lr,#4
;------------------------------------------------------------------------------
;- Function             : IRQ_Handler_Entry
;- Treatments           : IRQ Controller Interrupt Handler.
;- Called Functions     : AIC_IVR[interrupt] 
;------------------------------------------------------------------------------

PUBLIC IRQ_Handler_Entry

IRQ_Handler_Entry:
;- Manage Exception Entry
;- Adjust and save LR_irq in IRQ stack
            sub         lr, lr, #4
            stmfd       sp!, {lr}

;- Save SPSR need to be saved for nested interrupt
            mrs         r14, SPSR
            stmfd       sp!, {r14}

;- Save and r0 in IRQ stack
            stmfd       sp!, {r0}

;- Write in the IVR to support Protect Mode
;- No effect in Normal Mode
;- De-assert the NIRQ and clear the source in Protect Mode
            ldr         r14, =AT91C_BASE_AIC
            ldr         r0 , [r14, #AIC_IVR]
	        str         r14, [r14, #AIC_IVR]

;- Enable Interrupt and Switch in Supervisor Mode
            msr         CPSR_c, #ARM_MODE_SVC

;- Save scratch/used registers and LR in User Stack
            stmfd       sp!, { r1-r3, r12, r14}

;- Branch to the routine pointed by the AIC_IVR
            mov         r14, pc
            bx          r0

;- Restore scratch/used registers and LR from User Stack
            ldmia       sp!, { r1-r3, r12, r14}

;- Disable Interrupt and switch back in IRQ mode
            msr         CPSR_c, #I_BIT | ARM_MODE_IRQ

;- Mark the End of Interrupt on the AIC
            ldr         r14, =AT91C_BASE_AIC
            str         r14, [r14, #AIC_EOICR]

;- Restore SPSR_irq and r0 from IRQ stack
            ldmia       sp!, {r0}

;- Restore SPSR_irq and r0 from IRQ stack
            ldmia       sp!, {r14}
            msr         SPSR_cxsf, r14

;- Restore adjusted  LR_irq from IRQ stack directly in the PC
            ldmia       sp!, {pc}^
            ENDP

;--------------------
;- The reset handler
;--------------------
InitReset:
;------------------------------------------------------------------------------
;- Low level Init (APMC, AIC,  ....) by C function AT91F_LowLevelInit
;------------------------------------------------------------------------------
;- minimum  C initialization

            ldr     r13,=Top_Stack    ; temporary stack in internal Ram
$IF (ARM_MODE)
            EXTERN  CODE32 (AT91F_LowLevelInit?A)
            ldr	    r0,=AT91F_LowLevelInit?A
$ELSE
            EXTERN  CODE32 (AT91F_LowLevelInit?T)
            ldr	    r0,=AT91F_LowLevelInit?T
$ENDIF
            mov     lr, pc
            bx	    r0

;------------------------------------------------------------------------------
;- Stack Sizes Definition
;------------------------
;- Interrupt Stack requires 2 words x 8 priority level x 4 bytes when using
;- the vectoring. This assume that the IRQ_ENTRY/IRQ_EXIT macro are used.
;- The Interrupt Stack must be adjusted depending on the interrupt handlers.
;- Fast Interrupt not requires stack If in your application it required you must
;- be define here.
;- Other stacks are defined by default to save one word each.
;- The System stack size is not defined and is limited by the free internal
;- SRAM.
;- User stack size is not defined and is limited by the free external SRAM.
;------------------------------------------------------------------------------


;------------------------------------------------------------------------------
;- Setup the stack for each mode
;-------------------------------
                mov     r0, r13

;- Set up Fast Interrupt Mode and set FIQ Mode Stack
                msr     CPSR_c, #ARM_MODE_FIQ | I_BIT | F_BIT
;- Init the FIQ register
            	ldr     r8, =AT91C_BASE_AIC

;- Set up Interrupt Mode and set IRQ Mode Stack
                msr     CPSR_c, #ARM_MODE_IRQ | I_BIT | F_BIT
                mov     r13, r0                     ; Init stack IRQ
                sub     r0, r0, #IRQ_Stack_Size

;- Set up Supervisor Mode and set Supervisor Mode Stack
                msr     CPSR_c, #ARM_MODE_SVC
                mov     r13, r0                     ; Init stack Sup


;------------------------------------------------------------------------------
;- Branch on C code Main function (with interworking)
;----------------------------------------------------
;- Branch must be performed by an interworking call as either an ARM or Thumb
;- main C function must be supported. This makes the code not position-
;- independent. A Branch with link would generate errors
;------------------------------------------------------------------------------
// Enter the C code
                EXTERN  CODE32 (?C?INIT)
                LDR     R0,=?C?INIT
				LDR     LR,=exit  ; ARM Mode
                BX      R0

                ENDP

;------------------------------------------------------------------------------
;- Loop for ever
;---------------
;- End of application. Normally, never occur.
;- Could jump on Software Reset ( B 0x0 ).
;------------------------------------------------------------------------------
$IF (ARM_MODE)
exit?A          PROC    CODE32
exit:           B       exit 
                ENDP
$ELSE

exit?T         PROC    CODE16
exit:           B       exit 
                ENDP
$ENDIF
;------------------------------------------------------------------------------
;- Manage exception
;---------------
;- This module The exception must be ensure in ARM mode
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
;- Manage exception
;---------------
;- The exception must be ensure in ARM mode
;------------------------------------------------------------------------------
PUBLIC AT91F_Default_FIQ_handler?A
AT91F_Default_FIQ_handler?A         PROC    CODE32
PUBLIC AT91F_Default_FIQ_handler?T
AT91F_Default_FIQ_handler?T         
            b     AT91F_Default_FIQ_handler?A
            ENDP
            

PUBLIC AT91F_Default_IRQ_handler?A
AT91F_Default_IRQ_handler?A     PROC    CODE32     

PUBLIC AT91F_Default_IRQ_handler?T
AT91F_Default_IRQ_handler?T          

            b     AT91F_Default_IRQ_handler?A
            ENDP

PUBLIC AT91F_Spurious_handler?A
AT91F_Spurious_handler?A          PROC    CODE32
PUBLIC AT91F_Spurious_handler?T
AT91F_Spurious_handler?T         
            b     AT91F_Spurious_handler?A
            ENDP
    END

